import requests
import re
from typing import List, Tuple, Generator

# Exceptions
class InvalidAuthException(Exception):
    def __init__(self):
        super().__init__("Invalid login credentials")

class NotLoggedInException(Exception):
    def __init__(self):
        super().__init__("Not logged in")

class StudentNotFoundException(Exception):
    def __init__(self):
        super().__init__("Student not found")

class NotImplementedException(Exception):
    def __init__(self):
        super().__init__("Not implemented")

class StudentInfo(object):
    def __init__(self, json: dict, endpoint: str):
        self.json = json
        self.student_id = json["record"]["studentID"]
        self.name = (json["record"]["firstname"], json["record"]["lastname"])
        self.gender = json["record"]["gender"]
        self.dob = json["record"]["dob"]
        self.timetable = json["record"]["timetable"].split(" ")
        self.username = json["record"]["loginID"]
        self.age = json["age"]
        self.photo = endpoint + json["photoURL"]

        self.gaurdian_phone = json["record"]["guardianPhone"]
        self.gaurdian_email = json["record"]["guardianEmail"]

    def __str__(self):
        return str(self.json)

class CourseInfo(object):
    def __init__(self, code: str, teacher_fname: str, teacher_lname: str, period: int):
        self.code = code
        self.teacher_name = (teacher_fname, teacher_lname)
        self.period = period

    def __str__(self):
        return str({
            "code": self.code,
            "period": self.period,
            "teacher_name": self.teacher_name
        })

class TimeAndPlace(object):

    logged_in: bool = False
    
    def __init__(self, endpoint: str = "http://demo.iquark.ca:8009"):
        """Create a TimeAndPlace client

        Args:
            endpoint (str, optional): TimeAndPlace server endpoint. Defaults to "http://demo.iquark.ca:8009".
        """
        self.endpoint = endpoint

    def _checkLogin(self):
        if not self.logged_in:
            raise NotLoggedInException()

    def login(self, username: str, password: str):
        """Log in to TimeAndPlace

        Args:
            username (str): Application username
            password (str): Application password

        Raises:
            Exception: Thrown if CSRF data cannot be parsed from TimeAndPlace
            InvalidAuthException: Thrown if login details are invalid
        """

        # Build a session
        sess = requests.session()

        # Get needed cookies from server
        login_page = sess.get(f"{self.endpoint}/login").text

        # Determine the CSRF token
        try:
            self.csrf_token = re.findall(r'<input type="hidden" name="_token" value="([a-zA-Z0-9]*)">', login_page, re.M)[0]
        except IndexError as e:
            raise Exception("Remote server responded with unusual data. Please contact administrator")

        # Make actual login request
        resp = sess.post(f"{self.endpoint}/login", data={
            "username": username,
            "password": password,
            "_token":self.csrf_token
        })

        # Look for the error message
        if "These credentials do not match our records." not in resp.text:
            self.logged_in = True
            self.rsession = sess
        else:
            raise InvalidAuthException()

    def getStudentInfo(self, student_id: int) -> StudentInfo:
        """Get a student's info object

        Args:
            student_id (int): Student ID

        Raises:
            StudentNotFoundException: Thrown if the student does not exist

        Returns:
            StudentInfo: StudentInfo object
        """

        # Make a creds request
        student = requests.get(f"{self.endpoint}/studentsJSON/{student_id}").json()

        # Check for invalid ID
        if student["age"] == -1:
            raise StudentNotFoundException()

        # Create student object
        student_obj = StudentInfo(student, self.endpoint)
        return student_obj

    def getAllStudents(self) -> Generator[int, None, None]:
        """Get a list of all registered student IDs

        Returns:
            List[int]: List of student IDs
        """
        
        # Require auth
        self._checkLogin()

        # Get student listing
        students_raw = self.rsession.get(f"{self.endpoint}/allstudents").text

        # Parse out all student info
        students = re.findall(r'<tr><td>([0-9]*) \. \. \. ([a-zA-Z -]*), ([a-zA-Z -]*)<\/td><td>([a-zA-Z0-9]*)<\/td><\/tr>', students_raw, re.M)

        # Send data
        for student in students:
            yield student[0]

    def getAllCourses(self) -> List[CourseInfo]:
        """Get a list of all current courses

        Returns:
            List[CourseInfo]: List of courses
        """
        raise NotImplementedException()
    
    def signInStudentToTerminal(self, terminal_id: int, student_id: int):
        """Sign a student in to a terminal

        Args:
            terminal_id (int): Terminal ID
            student_id (int): Student ID

        Raises:
            StudentNotFoundException: Thrown if the student does not exist
        """

        # Make a toggle request
        status = requests.get(f"{self.endpoint}/terminals/{terminal_id}/toggleStudentID/{student_id}").json()

        # Check if the student was already signed in
        if status.get("status", "signed in") == "signed out":
            # Do signin again to fix the state
            requests.get(f"{self.endpoint}/terminals/{terminal_id}/toggleStudentID/{student_id}").json()

        elif status.get("status", "signed in") == "not found":
            raise StudentNotFoundException()

    def signOutStudentFromTerminal(self, terminal_id: int, student_id: int):
        """Sign a student out of a terminal

        Args:
            terminal_id (int): Terminal ID
            student_id (int): Student ID

        Raises:
            StudentNotFoundException: Thrown if the student does not exist
        """

        # Make a toggle request
        status = requests.get(f"{self.endpoint}/terminals/{terminal_id}/toggleStudentID/{student_id}").json()

        # Check if the student was already signed out
        if status.get("status", "signed in") == "signed in":
            # Do signin again to fix the state
            requests.get(f"{self.endpoint}/terminals/{terminal_id}/toggleStudentID/{student_id}").json()

        elif status.get("status", "signed in") == "not found":
            raise StudentNotFoundException()

    

    


from . import *
import argparse
import json
import sys
import terminaltables

def sign_in(args, client: TimeAndPlace):
    
    # Ensure we have needed args
    if not args.student or not args.terminal:
        print("Requires --student and --terminal to be set", file=sys.stderr)
        exit(1)

    # Make signin request
    try:
        client.signInStudentToTerminal(args.terminal, args.student)
    except StudentNotFoundException:
        print(f"Student {args.student} does not exist", file=sys.stderr)
        exit(1)

    print("Signed in")

def sign_out(args, client: TimeAndPlace):
    
    # Ensure we have needed args
    if not args.student or not args.terminal:
        print("Requires --student and --terminal to be set", file=sys.stderr)
        exit(1)

    # Make signout request
    try:
        client.signOutStudentFromTerminal(args.terminal, args.student)
    except StudentNotFoundException:
        print(f"Student {args.student} does not exist", file=sys.stderr)
        exit(1)
    
    print("Signed out")

def get_student_info(args, client: TimeAndPlace):
    
    # Ensure we have needed args
    if not args.student:
        print("Requires --student to be set", file=sys.stderr)
        exit(1)

    # Make lookup request
    student:dict
    try:
        student = client.getStudentInfo(args.student)
    except StudentNotFoundException:
        print(f"Student {args.student} does not exist", file=sys.stderr)
        exit(1)

    # Build user data table
    table = terminaltables.SingleTable([
        ["Name", f"{student.name[1]}, {student.name[0]}"],
        ["ID", f"{student.student_id}"],
        ["Username", f"{student.username}"],
        ["Gender", f"{student.gender}"],
        ["Birthdate", f"{student.dob}"],
        ["Age", f"{student.age}"],
        ["Photo", f"{student.photo}"],
        ["Gaurdian Email", f"{student.gaurdian_email}"],
        ["Gaurdian Phone", f"{student.gaurdian_phone}"],
        ["Courses", ", ".join(student.timetable)],
    ])

    print(table.table)
    

def get_student_ids(args, client: TimeAndPlace):
    
    # Ensure we have needed args
    if not args.username or not args.password:
        print("Requires --username and --password to be set", file=sys.stderr)
        exit(1)
    
    # Log in
    try:
        client.login(args.username, args.password)
    except InvalidAuthException:
        print("Invalid authentication info", file=sys.stderr)
        exit(1)

    # Make lookup request
    students = list(client.getAllStudents())
    students.sort()

    # Dump data
    for student in students:
        print(student)

# Command map
command_actions = {
    "signin": (sign_in, "Sign a student in to a terminal"),
    "signout": (sign_out, "Sign a student out form a terminal"),
    "student/info": (get_student_info, "Get JSON-formatted info about a student"),
    "students/ids": (get_student_ids, "Get a raw list of all valid student IDs"),
}

# Handle CLI args
ap = argparse.ArgumentParser(description="Use the \"help\" action for info on actions")

# Endpoints
ap.add_argument("-e", "--endpoint", help="Set a custom server address")

# Auth
ap.add_argument("--username", help="TimeAndPlace username", required=False)
ap.add_argument("--password", help="TimeAndPlace password", required=False)

# Flags
ap.add_argument("-s", "--student", help="Student ID", required=False)
ap.add_argument("-t", "--terminal", help="Terminal ID", required=False)

# Action
ap.add_argument("action", help="Action to run")

# Get args
args = ap.parse_args()

# Handle help action
if args.action == "help":
    # Build table data
    data = [[key, command_actions[key][1]] for key in command_actions]
    table = terminaltables.SingleTable(data)

    # Display table
    table.title = "Available actions"
    table.inner_heading_row_border = False
    print(table.table)

    # print("List of available actions:")
    # print("  Action\t\tDescription")

    # for key in command_actions:
    #     print(f"  {key}\t\t{command_actions[key][1]}")

    exit(0)

# Check that action is valid
if not args.action in command_actions:
    print("Action not valid. See help.", file=sys.stderr)
    exit(1)

# Create a new client
if args.endpoint:
    client = TimeAndPlace(endpoint = args.endpoint)
else:
    client = TimeAndPlace()

# Run action
command_actions[args.action][0](args, client)